import { json, urlencoded } from 'body-parser';
import compression from 'compression';
import cors from 'cors';
import express, { NextFunction, Request, Response } from 'express';
import helmet from 'helmet';
import {
  getDescriptor,
  ICoronerRequestOption,
  IdentityManager,
  IServiceDescriptor,
  listenDescriptor,
} from '..';
import { Logger } from '../log/logger';
import { MetricsStorage } from '../metrics/metricsStorage';
import { IAppOptions } from '../model/appOptions';
import { IConfiguration } from '../model/configuration';
import { ServiceConfiguration } from './serviceConfiguration';

export class BacktraceService<T extends {}> {
  /**
   * Create and startup express application with basic options.
   * By using this method, application will apply all
   * Backtrace service rules
   * @param name service name (required by service layer)
   * @param port service port (required by service layer)
   * @param opts optional application options
   * @param app  optional express app instance
   * By using this parameter you can create express app in different
   * framework and still be able to use Backtrace integrations
   */
  public static bootstrap<T extends {}>(
    name: string,
    port: number,
    opts?: IAppOptions,
    app?: express.Express,
  ): BacktraceService<T> {
    if (!port || !name) {
      throw new Error(`Invalid arguments. Name: ${name}, port: ${port}`);
    }
    const service = new BacktraceService<T>(name, port, opts, app);
    app = service.app;
    this.setupAuth(app);
    this.setupMiddlewares(app);
    this.setupServiceLayer(app, name, port, (ip) => {
      service.coronerdAddress.add(ip);
      if (opts && opts.coronerdCallback) {
        opts.coronerdCallback(ip);
      }
    });
    if (!name.startsWith('backtrace')) {
      name = `backtrace=${name}`;
    }
    process.title = name;

    app.on('close', () => {
      service._logger.warn('server is closing');
    });
    return service;
  }

  public static readConfiguration<T = {}>(
    name: string,
    port: number,
    path?: string,
    ignoreDescriptor?: boolean,
  ): IConfiguration<T> {
    const serviceConfiguration = new ServiceConfiguration<T>();
    if (!path) {
      path = serviceConfiguration.getProjectSettingsPath(name);
    }
    const configuration = serviceConfiguration.loadConfiguration(path);
    if (!ignoreDescriptor) {
      configuration.descriptor = getDescriptor(name, port);
    }
    return configuration;
  }

  /**
   * Setup popular Backtrace middlewares
   * @param app express app
   */
  public static setupMiddlewares(app: express.Express) {
    app.use(json());
    app.use(urlencoded({ extended: true }));
    app.use(compression());

    // Route hit by pingdom
    app.get('/', (_req, res) => {
      res.sendStatus(200);
    });

    app.use(
      (
        err: Error & {
          statusCode: number;
          response: { status: number; message: string };
        },
        req: Request,
        res: Response,
        next: NextFunction,
      ) => {
        // handle a situation when we already sent a response from onError method
        // we shouldn't do that and we should let error middleware handle that
        // but here we wan't to make sure our code won't break up
        // and we won't generate another error
        if (res.headersSent) {
          return;
        }

        const logger = Logger.getLogger();
        // Backtrace service ApiError (pending in CR) contains type statusCode that allows us
        // to determine what type of response we should return
        // in addition to that I recognize sometimes we retrieve status code from response.status
        //  property. This is a popular pattern for axios communication - check comments and hooks.

        if (err.statusCode || (err.response && err.response.status)) {
          const status = err.statusCode || err.response.status;
          const message = err.message || err.response.message;
          if (status === 500) {
            logger.error(err);
          } else {
            logger.warn(message);
          }
          res.status(status).send(message);
          return;
        }
        logger.error(err);
        res.sendStatus(500);
      },
    );
  }

  /**
   * Setup security middlewares
   * @param app express app
   */
  public static setupAuth(app: express.Express) {
    app.use(cors());
    app.options('*', cors());
    app.use(helmet.noCache());
    app.use(
      helmet({
        hsts: false,
      }),
    );
  }

  /**
   * Setup service layer integration
   * @param app express app
   */
  public static setupServiceLayer(
    app: express.Express,
    name?: string,
    port?: number,
    coronerdCallback?: (param: string) => void,
  ) {
    const logger = Logger.getLogger();
    let serviceDescriptor = this._service
      ? this._service._descriptor
      : undefined;

    // this case handle a situation when we didn't call backtraceService constr
    // here we want to read service descriptor from hard drive - if descr exists
    // we're using here a service name and port instead of service descriptor
    // to avoid extra steps required by service to setup service layer.
    if (!serviceDescriptor && name && port) {
      serviceDescriptor = getDescriptor(name, port);
    }

    if (!serviceDescriptor) {
      throw new Error('Descriptor is undefined');
    }

    const identityManager = this._service
      ? this._service.identityManager
      : new IdentityManager(serviceDescriptor);

    app.post(
      `/api/${serviceDescriptor.name}/service`,
      identityManager.serviceRequest({
        name: serviceDescriptor.name,
        secret: serviceDescriptor.secret,
        coronerdCallback,
      } as ICoronerRequestOption),
    );
  }
  private static _service: BacktraceService<any>;

  public readonly configuration!: IConfiguration<T>;
  public readonly coronerdAddress = new Set<string>();

  public identityManager!: IdentityManager;
  public readonly app: express.Express;
  public readonly metrics!: MetricsStorage;

  private readonly _logger: Logger;
  private _descriptor!: IServiceDescriptor;

  constructor(
    private readonly name: string,
    private readonly port: number,
    private readonly _opts?: IAppOptions,
    app?: express.Express,
  ) {
    this.app = app ? app : express();
    this.configuration = BacktraceService.readConfiguration<T>(
      this.name,
      this.port,
      this._opts && this._opts.confPath ? this._opts.confPath : undefined,
    );

    const logLevel =
      this.configuration.logger && this.configuration.logger.logLevel
        ? this.configuration.logger.logLevel
        : undefined;

    this._logger =
      this._opts && this._opts.logger
        ? this._opts.logger
        : new Logger(logLevel);

    if (
      this.configuration.backtrace &&
      this.configuration.backtrace.remoteLog !== false
    ) {
      Logger.initializeBacktrace(this.configuration.backtrace);
    }

    this.metrics = new MetricsStorage(
      this.configuration.metrics && this.configuration.metrics.uuid,
      this.configuration.metrics && this.configuration.metrics.secret,
      undefined,
      this.app,
    );

    this.prepareServiceLayer();
    BacktraceService._service = this;
  }

  public start() {
    if (!this.app) {
      this._logger.error(
        `Express application is not defined. Cannot start application`,
      );
      return;
    }
    if (!this._descriptor) {
      this._logger.error(
        `Descriptor object is not defined.
        Please execute setupServiceLayer method`,
      );
      return;
    }

    listenDescriptor(this._descriptor, this.app, this.configuration.ssl, () => {
      this._logger.info(
        `Application is Running. ${
          this._descriptor ? `Port: ${this._descriptor.port}` : ''
        }`,
      );
    });
  }

  private prepareServiceLayer() {
    if (!this._descriptor) {
      const descriptor = getDescriptor(this.name, this.port);
      this._descriptor = descriptor;
    }
    this.identityManager = new IdentityManager(this._descriptor);
  }
}
