const express = require('express');
const cors = require('cors');
const btoa = require('btoa');
const coronerLogin = require('../util/coronerLogin');
const { samlSuccess, samlFailure } = require('../util/metricsCache');
const sendReport = require('../util/sendReport');
const { service } = require('../util/service-layer');
const {
  fuzzyGetAuthenticatedCoronerd,
  rehydtrateToken,
} = require('../util/coronerdTokenCache');

const {
  getBPGUsers,
  createBPGUsers,
  getUniqueUsername,
} = require('../util/createUser');

const config = service.configuration;

const DEFAULT_SERVICE_URL = 'https://saml.backtrace.io';

const encode = str => encodeURIComponent(btoa(str));

const getCoronerdToken = (req, res, next) => {
  const host = req.user.tenant;

  const authenticatedCoronerd = fuzzyGetAuthenticatedCoronerd(host);

  const coronerdUrl = host.startsWith('https://') ? host : 'https://' + host;

  req.coronerdUrl = coronerdUrl;

  // Check if the coronerdToken exists in token cache
  if (
    service.identityManager.authenticated[authenticatedCoronerd] &&
    service.identityManager.authenticated[authenticatedCoronerd]._config
  ) {
    req.coronerdToken =
      service.identityManager.authenticated[
        authenticatedCoronerd
      ]._config.token;
    req.coronerdUniverse =
      service.identityManager.authenticated[
        authenticatedCoronerd
      ]._config.universe;
    next();
  } else {
    service.identityManager
      .loginCoronerd(coronerdUrl)
      .then(configuration => {
        req.coronerdToken = configuration.token;
        req.coronerdUniverse = configuration.universe;
        next();
      })
      .catch(error => {
        req.loginCoronerdError = error;
        next(); // continue on and use whitelist as a fallback
      });
  }
};

function parseResponseError(err) {
  let error;

  // This gives us the actual error from the POST request to Coronerd
  try {
    error = new Error(err.response.data.error.message);
  } catch (_error) {
    error = err;
  }

  return error;
}

// function to pull token from coronerd response to be used in redirectToConsole
function getTokenFromResponse(response) {
  if (response.data && response.data.response && response.data.response.token) {
    return response.data.response.token;
  } else {
    const error =
      response.error && response.error.message
        ? new Error(response.error.message)
        : new Error('failed to login');
    throw error;
  }
}

const getUserToken = (req, res, next) => {
  const host = req.user.tenant;
  const email = req.user.nameID;
  const token = req.coronerdToken;
  const universe = req.coronerdUniverse;
  const coronerdUrl = req.coronerdUrl;

  const captureError = error => {
    samlFailure(host);
    sendReport(error, {
      nameID: req.user.nameID,
      payload: req.payload,
      hostname: host,
    });
    next(error);
  };

  coronerLogin(coronerdUrl, email, token)
    .then(getTokenFromResponse)
    .then(loginToken => {
      req.loginToken = loginToken;
      req.coronerdUrl = coronerdUrl;
      next();
    })
    .catch(err => {
      let error = parseResponseError(err);

      switch (error.message) {
        // rehydrate token on failure and retry login else move on to error middleware
        case 'permission denied':
          rehydtrateToken(coronerdUrl)
            .then(config => coronerLogin(coronerdUrl, email, config.token))
            .then(getTokenFromResponse)
            .then(loginToken => {
              req.loginToken = loginToken;
              req.coronerdUrl = coronerdUrl;
              next();
            })
            .catch(captureError);
          break;
        case 'no such user':
          if (
            req.session.tenant &&
            req.session.tenant.settings &&
            req.session.tenant.settings.userProvisioning
          ) {
            getBPGUsers(coronerdUrl, token)
              .then(users => {
                const username = getUniqueUsername(users, email);
                const user = {
                  username,
                  email,
                  universe: universe.id,
                };
                return user;
              })
              .then(user => createBPGUsers(coronerdUrl, token, user))
              .then(_response => coronerLogin(coronerdUrl, email, token))
              .then(getTokenFromResponse)
              .then(loginToken => {
                req.loginToken = loginToken;
                req.coronerdUrl = coronerdUrl;
                next();
              })
              .catch(err => {
                const error = parseResponseError(err);
                captureError(error);
              });
          } else {
            captureError(error);
          }
          break;
        default:
          captureError(error);
          break;
      }
    });
};

// Method to redirect client back to Backtrace on successful login
const redirectToConsole = (req, res, next) => {
  const { coronerdUrl, loginToken } = req;
  const host = req.user.tenant;
  samlSuccess(host);
  res.redirect(`${coronerdUrl}/login?method=saml&token=${loginToken}`);
};

const chooseHost = (req, res, next) => {
  try {
    // If there exists a lastUniverse in the session then set that as a tenant
    if (req.session && req.session.lastUniverse && req.user) {
      req.user = Object.assign(req.user, {
        tenant: req.session.lastUniverse,
      });
      return next();
    }

    // TODO deprecate this function
    // if hostname is a key in the gateway object
    //   render chooseHost view
    // else
    //   go on to redirectToConsole function
    if (config.gateway && config.gateway[req.params.hostname]) {
      return res.render('chooseHost', {
        user: req.user,
        tenants: config.gateway[req.params.hostname].sort(),
      });
    }

    // if sessions are disabled / not working, use the hostname in the params
    req.user = Object.assign(req.user, {
      tenant: req.params.hostname,
    });

    next();
  } catch (error) {
    console.error(error);
    const hostname =
      req.session && req.session.lastUniverse
        ? req.session.lastUniverse
        : req.params.hostname;
    sendReport(error, {
      hostname,
      nameID: req.user.nameID,
      session: JSON.stringify(req.session),
    });
    next(new Error('Failure to reconcile host'));
  }
};

// Method to set last visited universe in session - used for chooseHost
const setLastUniverse = (req, res, next) => {
  req.session.lastUniverse = req.params.hostname;
  next();
};

const passportCb = (req, res, next) => {
  return (err, user, info) => {
    if (err) {
      const hostname = req.params.hostname;
      const attributes = {};

      if (user && user.nameID) {
        attributes.nameID = user.nameID;
      }

      samlFailure(hostname);

      sendReport(err, {
        nameID: user && user.nameID ? user.nameID : '',
        hostname,
      });

      return next(err);
    }

    if (!user) {
      return next(new Error('passport error: no user returned'));
    }

    req.user = user;
    next();
  };
};

const handleSamlError = (error, req, res, next) => {
  let context = encode('failed to login');
  let message = encode(error.message);

  // find hostname either in the session store or grab from the
  // '/saml/*' route params
  let hostname;
  if (req.session && req.session.lastUniverse) {
    hostname = req.session.lastUniverse;
  } else {
    hostname = req.params[0];
  }

  // IF the error message is 'no such user' && no user provisioning
  // THEN show this message
  if (
    error.message.includes('no such user') ||
    (req.payload && req.payload.includes('no such user'))
  ) {
    context = encode('No such user');
    message = encode(
      `If this is the first time logging in, please create a Backtrace account first via "Create Account" on the main login page. If you still have trouble logging in, please contact ${
        req.session.tenant && req.session.tenant.adminContact
          ? req.session.tenant.adminContact
          : 'your Backtrace administrator'
      }.`
    );
  }

  res.redirect(
    `https://${hostname}/error?context=${context}&message=${message}`
  );
};

module.exports = passports => {
  const router = express.Router();
  // GET request sent from Console's login page when clicking
  // 'Sign in with SSO' button
  router.get(
    '/login/:hostname',
    setLastUniverse,
    passports.attach(),
    passports.middleware('initialize'),
    passports.middleware('session'),
    passports.middleware('authenticate', 'saml')
  );

  // POST sent IdP back to SAML service for authenticating login with
  // Passport SAML strategy
  router.post(
    '/:hostname',
    passports.attach(),
    passports.middleware('initialize'),
    (req, res, next) => {
      const cb = passportCb(req, res, next);
      passports.middleware('authenticate', 'saml', cb)(req, res, next);
    },
    chooseHost,
    getCoronerdToken,
    getUserToken,
    redirectToConsole
  );

  // handler for all /saml/* route errors
  router.use('/*', handleSamlError);

  return router;
};
